@@ -1,23 +1,30 @@
 import abc
 import base64
+import cgi
 import gzip
 import json
 import logging
 import os
 import re
+from datetime import datetime
 from enum import Enum
-from typing import ClassVar, Dict, List, Optional, Type, Union, get_type_hints
+from typing import ClassVar, Dict, List, Optional, Tuple, Type, Union, get_type_hints
 
 import gcsfs
 import humanize
 import pendulum
-from pydantic import BaseModel, Field, HttpUrl, validator
+from google.cloud import storage
+from jinja2 import Environment, StrictUndefined, select_autoescape
+from pydantic import BaseModel, Field, HttpUrl, constr, validator
 from pydantic.class_validators import root_validator
 from pydantic.tools import parse_obj_as
-from requests import Request
+from requests import Request, Session
 from typing_extensions import Annotated, Literal
 
-from .config import get_bucket, is_cloud, is_development, require_pipeline
+from .config import get_bucket, is_cloud, require_pipeline
+
+JSONL_EXTENSION = ".jsonl"
+JSONL_GZIP_EXTENSION = f"{JSONL_EXTENSION}.gz"
 
 
 def get_fs(gcs_project="", **kwargs):
@@ -92,11 +99,9 @@ def make_name_bq_safe(name: str):
     return str.lower(re.sub("[^\w]", "_", name))  # noqa: W605
 
 
-def prefix_bucket(bucket):
-    # TODO: use once we're in python 3.9+
-    # bucket = bucket.removeprefix("gs://")
-    bucket = bucket.replace("gs://", "")
-    return f"gs://test-{bucket}" if is_development() else f"gs://{bucket}"
+AIRTABLE_BUCKET = os.getenv("CALITP_BUCKET__AIRTABLE")
+SCHEDULE_RAW_BUCKET = os.getenv("CALITP_BUCKET__GTFS_SCHEDULE_RAW")
+RT_RAW_BUCKET = os.getenv("CALITP_BUCKET__GTFS_RT_RAW")
 
 
 PARTITIONED_ARTIFACT_METADATA_KEY = "PARTITIONED_ARTIFACT_METADATA"
@@ -106,22 +111,20 @@ PartitionType = Union[str, int, pendulum.DateTime, pendulum.Date, pendulum.Time]
 PARTITION_SERIALIZERS = {
     str: str,
     int: str,
+    pendulum.Date: lambda d: d.to_date_string(),
     pendulum.DateTime: lambda dt: dt.to_iso8601_string(),
-    pendulum.Date: lambda d: d.strftime("%Y-%m-%d"),
-    pendulum.Time: lambda t: t.format("HH:mm:ss"),
 }
 
 PARTITION_DESERIALIZERS = {
     str: str,
     int: int,
-    pendulum.DateTime: lambda s: pendulum.parse(s, exact=True),
     pendulum.Date: lambda s: pendulum.parse(s, exact=True),
-    pendulum.Time: lambda s: pendulum.parse(s, exact=True),
+    pendulum.DateTime: lambda s: pendulum.parse(s, exact=True),
 }
 
 
 def partition_map(path) -> Dict[str, PartitionType]:
-    return {key: value for key, value in re.findall(r"/(\w+)=([\w\-:=]+)(?=/)", path.lower())}
+    return {key: value for key, value in re.findall(r"/(\w+)=([\w\-:=+.]+)(?=/)", path)}
 
 
 # class GTFSFeedType(str, Enum):
@@ -135,21 +138,57 @@ class GTFSFeedType(str, Enum):
     trip_updates = "trip_updates"
     vehicle_positions = "vehicle_positions"
 
+    @property
+    def is_rt(self) -> bool:
+        if self == GTFSFeedType.schedule:
+            return False
+        if self in (
+            GTFSFeedType.service_alerts,
+            GTFSFeedType.trip_updates,
+            GTFSFeedType.vehicle_positions,
+        ):
+            return True
+        raise RuntimeError("we should not get here")
+
 
 class AirtableGTFSDataRecord(BaseModel):
     name: str
     uri: Optional[str]
-    data: GTFSFeedType
+    data: Optional[GTFSFeedType]
     data_quality_pipeline: Optional[bool]
     schedule_to_use_for_rt_validation: Optional[List[str]]
     auth_query_param: Dict[str, str] = {}
-    # TODO: add auth_headers when available in Airtable!
 
     class Config:
         extra = "allow"
 
+    # TODO: this is a bit hacky but we need this until we split off auth query params from the URI itself
+    @root_validator(pre=True, allow_reuse=True)
+    def parse_query_params(cls, values):
+        if values.get("uri"):
+            jinja_pattern = r"(?P<param_name>\w+)={{\s*(?P<param_lookup_key>\w+)\s*}}&?"
+            match = re.search(jinja_pattern, values["uri"])
+            if match:
+                values["auth_query_param"] = {match.group("param_name"): match.group("param_lookup_key")}
+                values["uri"] = re.sub(jinja_pattern, "", values["uri"])
+            elif "api.511.org" in values["uri"]:
+                values["auth_query_param"] = {"api_key": "MTC_511_API_KEY"}
+        return values
+
+    @validator("uri", pre=True, allow_reuse=True)
+    def handle_remaining_jinja(cls, v):
+        if v:
+            data = {
+                "GRAAS_SERVER_URL": os.environ["GRAAS_SERVER_URL"],
+            }
+            return Environment(autoescape=select_autoescape(), undefined=StrictUndefined).from_string(v).render(**data)
+        return v
+
     @validator("data", pre=True, allow_reuse=True)
     def convert_feed_type(cls, v):
+        if not v:
+            return None
+
         if "schedule" in v.lower():
             return GTFSFeedType.schedule
         elif "vehicle" in v.lower():
@@ -158,6 +197,7 @@ class AirtableGTFSDataRecord(BaseModel):
             return GTFSFeedType.trip_updates
         elif "alerts" in v.lower():
             return GTFSFeedType.service_alerts
+
         return v
 
     @property
@@ -165,8 +205,18 @@ class AirtableGTFSDataRecord(BaseModel):
         # TODO: implement me
         raise NotImplementedError
 
+    # TODO: this should actually rely on airtable data!
     @property
     def auth_header(self) -> Dict[str, str]:
+        if self.uri:
+            if "goswift.ly" in self.uri:
+                return {
+                    "authorization": "SWIFTLY_AUTHORIZATION_KEY_CALITP",
+                }
+            if "west-hollywood" in self.uri:
+                return {
+                    "x-umo-iq-api-key": "WEHO_RT_KEY",
+                }
         return {}
 
     @property
@@ -199,7 +249,7 @@ class PartitionedGCSArtifact(BaseModel, abc.ABC):
     download of a given data source.
     """
 
-    filename: str
+    filename: constr(strip_whitespace=True)
 
     class Config:
         json_encoders = {
@@ -211,6 +261,11 @@ class PartitionedGCSArtifact(BaseModel, abc.ABC):
     def bucket(self) -> str:
         """Bucket name"""
 
+    @validator("bucket", check_fields=False, allow_reuse=True)
+    def bucket_exists(cls, v):
+        if not storage.Client().bucket(v).exists():
+            raise ValueError(f"bucket {v} does not exist")
+
     @property
     @abc.abstractmethod
     def table(self) -> str:
@@ -235,12 +290,27 @@ class PartitionedGCSArtifact(BaseModel, abc.ABC):
 
     @root_validator(allow_reuse=True)
     def check_partitions(cls, values):
+        # TODO: this isn't great but some partition_names are properties
+        if isinstance(cls.partition_names, property):
+            return values
         cls_properties = [name for name in dir(cls) if isinstance(getattr(cls, name), property)]
         missing = [name for name in cls.partition_names if name not in values and name not in cls_properties]
         if missing:
             raise ValueError(f"all partition names must exist as fields or properties; missing {missing}")
         return values
 
+    # TODO: these are repetitive
+    @property
+    def name(self):
+        return os.path.join(
+            self.table,
+            *[
+                f"{name}={PARTITION_SERIALIZERS[type(getattr(self, name))](getattr(self, name))}"
+                for name in self.partition_names
+            ],
+            self.filename,
+        )
+
     @property
     def path(self):
         return os.path.join(
@@ -253,55 +323,79 @@ class PartitionedGCSArtifact(BaseModel, abc.ABC):
             self.filename,
         )
 
-    def save_content(self, fs: gcsfs.GCSFileSystem, content: bytes, exclude=None):
-        logging.info(f"saving {humanize.naturalsize(len(content))} to {self.path}")
-        fs.pipe(path=self.path, value=content)
-        fs.setxattrs(
-            path=self.path,
-            # This syntax seems silly but it's so we pass the _value_ of PARTITIONED_ARTIFACT_METADATA_KEY
-            **{PARTITIONED_ARTIFACT_METADATA_KEY: self.json(exclude=exclude)},
-        )
+    def save_content(self, content: bytes, exclude=None, fs: gcsfs.GCSFileSystem = None, client: storage.Client = None):
+        if (fs is None) == (client is None):
+            raise TypeError("must provide a gcsfs file system OR a storage client")
+
+        if fs:
+            logging.info(f"saving {humanize.naturalsize(len(content))} to {self.path}")
+            fs.pipe(path=self.path, value=content)
+            fs.setxattrs(
+                path=self.path,
+                # This syntax seems silly but it's so we pass the _value_ of PARTITIONED_ARTIFACT_METADATA_KEY
+                **{PARTITIONED_ARTIFACT_METADATA_KEY: self.json(exclude=exclude)},
+            )
+
+        if client:
+            logging.info(f"saving {humanize.naturalsize(len(content))} to {self.bucket} {self.name}")
+            blob = storage.Blob(
+                name=self.name,
+                bucket=client.bucket(self.bucket.replace("gs://", "")),
+            )
+
+            blob.upload_from_string(
+                data=content,
+                content_type="application/octet-stream",
+                client=client,
+            )
+
+            blob.metadata = {PARTITIONED_ARTIFACT_METADATA_KEY: self.json(exclude=exclude)}
+            blob.patch()
 
 
 def fetch_all_in_partition(
     cls: Type[PartitionedGCSArtifact],
     fs: gcsfs.GCSFileSystem,
-    bucket: str,
-    table: str,
     partitions: Dict[str, PartitionType],
+    bucket: str = None,
+    table: str = None,
+    verbose=False,
+    progress=False,
 ) -> List[Type[PartitionedGCSArtifact]]:
-    path = "/".join(
+    if not bucket:
+        bucket = cls.bucket
+
+        if not isinstance(bucket, str):
+            raise TypeError("must either pass bucket, or the bucket must resolve to a string")
+
+    if not table:
+        table = cls.table
+
+        if not isinstance(table, str):
+            raise TypeError("must either pass table, or the table must resolve to a string")
+
+    prefix = "/".join(
         [
-            bucket,
             table,
             *[f"{key}={value}" for key, value in partitions.items()],
+            "",
         ]
     )
-    files = parse_obj_as(
-        GCSObjectInfoList,
-        [v for _, _, files in list(fs.walk(path, detail=True)) for v in files.values()],
-    )
-    return [
-        parse_obj_as(cls, json.loads(fs.getxattr(file.name, PARTITIONED_ARTIFACT_METADATA_KEY)))
-        for file in files.__root__
-    ]
+    if verbose:
+        print(f"listing all files in {bucket}/{prefix}")
+    client = storage.Client()
+    files = client.list_blobs(bucket.removeprefix("gs://"), prefix=prefix, delimiter=None)
+    return [parse_obj_as(cls, json.loads(file.metadata[PARTITIONED_ARTIFACT_METADATA_KEY])) for file in files]
 
 
 class ProcessingOutcome(BaseModel, abc.ABC):
     success: bool
     exception: Optional[Exception]
-    input_record: BaseModel
 
     class Config:
         arbitrary_types_allowed = True
         json_encoders = {Exception: lambda e: str(e)}
 
-    # TODO: is this really useful?
-    @property
-    @abc.abstractmethod
-    def input_type(self) -> Type[PartitionedGCSArtifact]:
-        """The input type that was processed to produce this outcome."""
-
 
 class GCSBaseInfo(BaseModel, abc.ABC):
     bucket: str
@@ -357,6 +451,7 @@ GCSObjectInfoList.update_forward_refs()
 # TODO: we could probably pass this a class
 def get_latest_file(table_path: str, partitions: Dict[str, Type[PartitionType]]) -> GCSFileInfo:
     fs = get_fs()
+    fs.invalidate_cache()
     directory = GCSDirectoryInfo(**fs.info(table_path))
 
     for key, typ in partitions.items():
@@ -381,99 +476,174 @@ def get_latest_file(table_path: str, partitions: Dict[str, Type[PartitionType]])
 
 
 class AirtableGTFSDataExtract(PartitionedGCSArtifact):
-    bucket: ClassVar[str] = prefix_bucket("gs://calitp-airtable")
+    bucket: ClassVar[str] = AIRTABLE_BUCKET
     table: ClassVar[str] = "california_transit__gtfs_datasets"
-    partition_names: ClassVar[List[str]] = ["dt", "time"]
-    dt: pendulum.Date
-    time: pendulum.Time
+    partition_names: ClassVar[List[str]] = ["dt", "ts"]
+    ts: pendulum.DateTime
     records: List[AirtableGTFSDataRecord]
 
+    @property
+    def dt(self) -> pendulum.Date:
+        return self.ts.date()
+
     # TODO: this should probably be abstracted somewhere... it's useful in lots of places, probably
     @classmethod
     def get_latest(cls) -> "AirtableGTFSDataExtract":
         # TODO: this concatenation should live on the abstract base class probably
         latest = get_latest_file(
             cls.bucket_table(),
+            # TODO: this doesn't pick up the type hint of dt since it's a property; it's fine as a string but we should fix
             partitions={name: get_type_hints(cls).get(name, str) for name in cls.partition_names},
         )
 
         logging.info(f"identified {latest.name} as the most recent extract of gtfs datasets")
 
         with get_fs().open(latest.name, "rb") as f:
-            content = gzip.decompress(f.read()).decode()
-
-        records = [AirtableGTFSDataRecord(**json.loads(row)) for row in content.splitlines()]
+            content = gzip.decompress(f.read())
 
         return AirtableGTFSDataExtract(
-            records=records,
             filename=latest.filename,
-            dt=pendulum.parse(latest.partition["dt"], exact=True),
-            time=pendulum.parse(latest.partition["time"], exact=True),
+            ts=pendulum.parse(latest.partition["ts"], exact=True),
+            records=[AirtableGTFSDataRecord(**json.loads(row)) for row in content.decode().splitlines()],
         )
 
 
 class GTFSFeedExtractInfo(PartitionedGCSArtifact):
-    # TODO: this should check whether the bucket exists https://stackoverflow.com/a/65628273
-    # TODO: this should be named `gtfs-raw` _or_ we make it dynamic
-    bucket: ClassVar[str] = prefix_bucket("gs://calitp-gtfs-schedule-raw")
-    partition_names: ClassVar[List[str]] = ["dt", "base64_url", "time"]
+    partition_names: ClassVar[List[str]] = ["dt", "base64_url", "ts"]
     config: AirtableGTFSDataRecord
     response_code: int
     response_headers: Optional[Dict[str, str]]
-    download_time: pendulum.DateTime
+    ts: pendulum.DateTime
+
+    @validator("ts", allow_reuse=True)
+    def coerce_ts(cls, v):
+        if isinstance(v, datetime):
+            return pendulum.instance(v)
+        return v
 
     @property
-    def table(self) -> GTFSFeedType:
+    def bucket(self) -> str:
+        if self.config.data.is_rt:
+            return RT_RAW_BUCKET
+        return SCHEDULE_RAW_BUCKET
+
+    @property
+    def table(self) -> str:
         return self.config.data
 
+    @property
+    def partition_names(self) -> List[str]:
+        if self.config.data.is_rt:
+            return ["dt", "hour", "ts", "base64_url"]
+        return ["dt", "base64_url", "ts"]
+
     @property
     def dt(self) -> pendulum.Date:
-        return self.download_time.date()
+        return self.ts.date()
+
+    @property
+    def hour(self) -> pendulum.DateTime:
+        return self.ts.replace(minute=0, second=0, microsecond=0)
 
     @property
     def base64_url(self) -> str:
         return self.config.base64_encoded_url
 
     @property
-    def time(self) -> pendulum.Time:
-        return self.download_time.time()
-
-
-class AirtableGTFSDataRecordProcessingOutcome(ProcessingOutcome):
-    input_type: ClassVar[Type[PartitionedGCSArtifact]] = GTFSFeedExtractInfo
-    extract: Optional[GTFSFeedExtractInfo]
-
-
-class DownloadFeedsResult(PartitionedGCSArtifact):
-    bucket: ClassVar[str] = prefix_bucket("gs://calitp-gtfs-schedule-raw")
-    table: ClassVar[str] = "download_schedule_feed_results"
-    partition_names: ClassVar[List[str]] = ["dt", "time"]
-    start_time: pendulum.DateTime
-    end_time: pendulum.DateTime
-    outcomes: List[AirtableGTFSDataRecordProcessingOutcome]
+    def timestamped_filename(self):
+        """
+        Used for RT validation; it's faster to download and validate many files at once, and we need to identify
+        them on disk.
+        """
+        return str(self.filename) + self.ts.strftime("__%Y-%m-%dT%H:%M:%SZ")
 
     @property
-    def dt(self) -> pendulum.Date:
-        return self.start_time.date()
+    def schedule_extract(self) -> "GTFSFeedExtractInfo":
+        if self.config.data == GTFSFeedType.schedule:
+            raise TypeError("cannot call schedule_extract on a schedule extract")
 
-    @property
-    def time(self) -> pendulum.Time:
-        return self.start_time.time()
-
-    @property
-    def successes(self) -> List[AirtableGTFSDataRecordProcessingOutcome]:
-        return [outcome for outcome in self.outcomes if outcome.success]
-
-    @property
-    def failures(self) -> List[AirtableGTFSDataRecordProcessingOutcome]:
-        return [outcome for outcome in self.outcomes if not outcome.success]
-
-    # TODO: I dislike having to exclude the records here
-    #   I need to figure out the best way to have a single type represent the "metadata" of
-    #   the content as well as the content itself
-    def save(self, fs):
-        self.save_content(
-            fs=fs,
-            content="\n".join(o.json() for o in self.outcomes).encode(),
-            exclude={"outcomes"},
+        file = get_latest_file(
+            GTFSFeedExtractInfo.bucket_table(),
+            partitions={name: get_type_hints(self).get(name, str) for name in self.partition_names},
         )
+        return parse_obj_as(
+            GTFSFeedExtractInfo, json.loads(get_fs().getxattr(file.name, PARTITIONED_ARTIFACT_METADATA_KEY))
+        )
+
+
+def download_feed(
+    record: AirtableGTFSDataRecord,
+    auth_dict: Dict,
+    ts: pendulum.DateTime,
+    default_filename="feed",
+) -> Tuple[GTFSFeedExtractInfo, bytes]:
+    parse_obj_as(HttpUrl, record.uri)
+
+    s = Session()
+    r = s.prepare_request(record.build_request(auth_dict))
+    resp = s.send(r)
+    resp.raise_for_status()
+
+    disposition_header = resp.headers.get("content-disposition", resp.headers.get("Content-Disposition"))
+
+    if disposition_header:
+        if disposition_header.startswith("filename="):
+            # sorry; cgi won't parse unless it's prefixed with the disposition type
+            disposition_header = f"attachment; {disposition_header}"
+        _, params = cgi.parse_header(disposition_header)
+        disposition_filename = params.get("filename")
+    else:
+        disposition_filename = None
+
+    filename = (
+        disposition_filename or (os.path.basename(resp.url) if resp.url.endswith(".zip") else None) or default_filename
+    )
+
+    extract = GTFSFeedExtractInfo(
+        filename=filename,
+        config=record,
+        response_code=resp.status_code,
+        response_headers=resp.headers,
+        ts=ts,
+    )
+
+    return extract, resp.content
+
+
+if __name__ == "__main__":
+    # just some useful testing stuff
+    # for record in AirtableGTFSDataExtract.get_latest().records:
+    #     if record.uri and "appspot" in record.uri:
+    #         print(record)
+    yesterday_noon = pendulum.yesterday("UTC").replace(minute=0, second=0, microsecond=0)
+    print(
+        len(
+            fetch_all_in_partition(
+                cls=GTFSFeedExtractInfo,
+                fs=get_fs(),
+                partitions={
+                    "dt": yesterday_noon.date(),
+                    "hour": yesterday_noon,
+                },
+                bucket=RT_RAW_BUCKET,
+                table=GTFSFeedType.vehicle_positions,
+                verbose=True,
+                progress=True,
+            )
+        )
+    )
+    print(
+        len(
+            fetch_all_in_partition(
+                cls=GTFSFeedExtractInfo,
+                fs=get_fs(),
+                partitions={
+                    "dt": pendulum.parse("2022-08-12", exact=True),
+                },
+                bucket=SCHEDULE_RAW_BUCKET,
+                table=GTFSFeedType.schedule,
+                verbose=True,
+                progress=True,
+            )
+        )
+    )
